from decimal import Decimal
from typing import Optional, Union


def is_blank(text: Optional[Union[str, int, float, Decimal]]) -> bool:
	if text is None:
		return True
	text = str(text)
	if len(text) == 0:
		return True
	if len(text.strip()) == 0:
		return True
	return False


def is_not_blank(text: Optional[str]) -> bool:
	return not is_blank(text)



from enum import Enum
from logging import DEBUG, ERROR, FATAL, Formatter, getLogger, handlers, INFO, NOTSET, StreamHandler, WARNING
from sys import stdout

from pydantic import BaseSettings
from pythonjsonlogger.jsonlogger import JsonFormatter


class LogLevel(str, Enum):
	FATAL = 'FATAL',
	ERROR = 'ERROR',
	WARN = 'WARN',
	INFO = 'INFO',
	DEBUG = 'DEBUG'


class LoggerSettings(BaseSettings):
	LOGGER_LEVEL: LogLevel = LogLevel.ERROR
	LOGGER_TO_FILE: bool = False
	LOGGER_FILE: str = 'temp/rotating.log'
	LOGGER_FILE_SIZE: int = 10242880
	LOGGER_FILE_BACKUP_COUNT: int = 5
	LOGGER_FILE_ENCODING: str = 'utf-8'
	LOGGER_JSON_FORMAT: bool = False
	# noinspection SpellCheckingInspection
	LOGGER_FORMAT: str = '%(asctime)s - %(process)d - %(threadName)s - %(name)s - %(levelname)s - %(message)s'

	class Config:
		env_file = '.env'
		env_file_encoding = 'utf-8'
		case_sensitive = True


def get_logger_level(level: LogLevel) -> int:
	if level == LogLevel.FATAL:
		return FATAL
	elif level == LogLevel.ERROR:
		return ERROR
	elif level == LogLevel.WARN:
		return WARNING
	elif level == LogLevel.INFO:
		return INFO
	elif level == LogLevel.DEBUG:
		return DEBUG
	else:
		return ERROR


def init_log():
	settings = LoggerSettings()
	logger = getLogger()
	logger.setLevel(NOTSET)

	logger_level = get_logger_level(settings.LOGGER_LEVEL)

	# Add stdout handler, with level INFO
	console = StreamHandler(stdout)
	console.setLevel(logger_level)
	formatter = Formatter(settings.LOGGER_FORMAT)
	console.setFormatter(formatter)
	logger.addHandler(console)

	# Add file rotating handler
	if settings.LOGGER_TO_FILE:
		file_log_handler = handlers.RotatingFileHandler(
			filename=settings.LOGGER_FILE,
			maxBytes=settings.LOGGER_FILE_SIZE,
			backupCount=settings.LOGGER_FILE_BACKUP_COUNT,
			encoding=settings.LOGGER_FILE_ENCODING)
		file_log_handler.setLevel(logger_level)
		if settings.LOGGER_JSON_FORMAT:
			formatter = JsonFormatter(settings.LOGGER_FORMAT)
		else:
			formatter = Formatter(settings.LOGGER_FORMAT)
		file_log_handler.setFormatter(formatter)
		logger.addHandler(file_log_handler)

	logger.info(f'Logger settings[{settings.dict()}].')



from .array_helper import ArrayHelper
from .datetime_helper import DateTimeConstants, DateTimeEncoder, get_current_time_in_seconds, get_day_of_month, \
	get_day_of_week, get_half_year, get_month, get_quarter, get_week_of_month, get_week_of_year, get_year, is_date, \
	is_date_or_time_instance, is_date_plus_format, is_datetime, is_time, month_diff, to_previous_month, \
	to_previous_week, to_yesterday, translate_date_format_to_memory, truncate_time, try_to_date, try_to_time, year_diff
from .json_helper import serialize_to_json
from .logger import init_log
from .numeric_helper import is_decimal, is_numeric_instance, try_to_decimal
from .string_helper import is_blank, is_not_blank
from .value_expression import equals_date, equals_decimal, equals_time, greater_or_equals_date, \
	greater_or_equals_decimal, greater_or_equals_time, is_empty, is_not_empty, less_or_equals_date, \
	less_or_equals_decimal, less_or_equals_time, value_equals, value_not_equals

init_log()



from __future__ import annotations

from typing import Any, Callable, List, Optional, Tuple

ArrayPredicate = Callable[[Any], bool]
ArrayFind = Callable[[Any], Tuple[bool, Any]]
ArrayTransform = Callable[[Any], Any]
ArrayWithIndexTransform = Callable[[Any, int], Any]
ArrayAction = Callable[[Any], None]
ArrayWithIndexAction = Callable[[Any, int], None]
ArrayCompare = Callable[[Any, Any], bool]
ArrayReduce = Callable[[Any, Any], Any]


def equals(a: Any, b: Any) -> bool:
	return a == b


class ArrayHelper:
	def __init__(self, a_list: Optional[list]):
		if a_list is None:
			self.aList = []
		else:
			self.aList = a_list

	def flatten(self, level: int = 1) -> ArrayHelper:
		new_list: list = []
		for an_element in self.aList:
			if isinstance(an_element, list):
				for a_sub_element in an_element:
					new_list.append(a_sub_element)
			else:
				new_list.append(an_element)

		level = level - 1
		if level <= 0:
			return ArrayHelper(new_list)
		else:
			return ArrayHelper(new_list).flatten(level)

	def to_list(self):
		return self.aList

	def to_map(self, as_key: ArrayTransform, as_value: ArrayTransform = lambda x: x) -> dict[Any, Any]:
		a_dict = {}
		for an_element in self.aList:
			key = as_key(an_element)
			value = as_value(an_element)
			a_dict[key] = value
		return a_dict

	def copy(self) -> ArrayHelper:
		return ArrayHelper(list(self.aList))

	def each(self, func: ArrayAction) -> ArrayHelper:
		"""
		apply given function to each element, and return myself
		"""
		for an_element in self.aList:
			func(an_element)
		return self

	def each_with_index(self, func: ArrayWithIndexAction) -> ArrayHelper:
		"""
		apply given function to each element, and return myself
		"""
		for index, an_element in enumerate(self.aList):
			func(an_element, index)
		return self

	def find(self, func: ArrayPredicate) -> Optional[Any]:
		"""
		find element which satisfies the given predicate function, returns None when not found
		"""
		for an_element in self.aList:
			if func(an_element):
				return an_element
		return None

	def filter(self, func: ArrayPredicate) -> ArrayHelper:
		"""
		pick elements which satisfies the given predicate function
		"""
		new_list: list = []
		for an_element in self.aList:
			if func(an_element):
				new_list.append(an_element)
		return ArrayHelper(new_list)

	def some(self, func: ArrayPredicate) -> bool:
		"""
		return true when at least one element satisfies the given predicate function
		"""
		for an_element in self.aList:
			if func(an_element):
				return True
		return False

	def every(self, func: ArrayPredicate) -> bool:
		"""
		return true when every element satisfies the given predicate function
		"""
		for an_element in self.aList:
			if not func(an_element):
				return False
		return True

	def map(self, func: ArrayTransform) -> ArrayHelper:
		"""
		transform each element by given transform function
		"""
		new_list: list = []
		for an_element in self.aList:
			new_list.append(func(an_element))
		return ArrayHelper(new_list)

	def map_with_index(self, func: ArrayWithIndexTransform) -> ArrayHelper:
		"""
		transform each element by given transform function
		"""
		new_list: list = []
		for index, an_element in enumerate(self.aList):
			new_list.append(func(an_element, index))
		return ArrayHelper(new_list)

	def reduce(self, func: ArrayReduce, accumulator: Optional[Any]) -> Any:
		if len(self.aList) == 0:
			return accumulator

		if accumulator is None:
			accumulator = self.aList[0]
			rest = self.aList[1:]
		else:
			rest = self.aList
		for an_element in rest:
			accumulator = func(accumulator, an_element)
		return accumulator

	# noinspection DuplicatedCode
	def distinct(self, func: Optional[ArrayCompare] = None) -> ArrayHelper:
		"""
		remove elements duplicated which satisfies given compare function
		"""
		if func is None:
			return ArrayHelper(list(set(self.aList)))

		new_list: list = []
		for an_element in self.aList:
			found: bool = False
			for an_existing_element in new_list:
				if func(an_element, an_existing_element):
					found = True
					break
			if not found:
				new_list.append(an_element)
		return ArrayHelper(new_list)

	def difference(self, another_list: Optional[list], func: Optional[ArrayCompare] = equals) -> ArrayHelper:
		"""
		pick elements which included in self but not in another
		"""
		if another_list is None or len(another_list) == 0:
			return self

		# noinspection DuplicatedCode
		new_list: list = []
		for an_element in self.aList:
			found: bool = False
			for an_another_element in another_list:
				if func(an_element, an_another_element):
					found = True
					break
			if not found:
				new_list.append(an_element)
		return ArrayHelper(new_list)

	def first(
			self,
			find: ArrayFind = lambda x: (True, x) if x is not None else (False, None)
	) -> Optional[Any]:
		"""
		first one of array which match the found function after transform_to function
		"""
		for an_element in self.aList:
			found, value = find(an_element)
			if found:
				return value
		return None

	def grab(self, *elements: Any) -> ArrayHelper:
		for an_element in elements:
			self.aList.append(an_element)
		return self

	def group_by(self, group: ArrayTransform) -> dict[Any, List[Any]]:
		a_dict = {}
		for an_element in self.aList:
			key = group(an_element)
			existing: list = a_dict.get(key)
			if existing is not None:
				existing.append(an_element)
			else:
				a_dict[key] = [an_element]
		return a_dict

	def join(self, separator: str) -> str:
		new_list: list = []
		for an_element in self.aList:
			if an_element is not None:
				new_list.append(str(an_element))
		return separator.join(new_list)

	def size(self) -> int:
		return len(self.aList)



from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, List, Optional, Tuple, Union

from .datetime_helper import is_date, is_time, truncate_time
from .numeric_helper import is_decimal


def is_empty(value: Any) -> bool:
	"""
	empty means given value is one of none, zero length string, empty list, empty dict
	"""
	if value is None:
		return True
	elif isinstance(value, str):
		return len(value) == 0
	elif isinstance(value, list):
		return len(value) == 0
	elif isinstance(value, dict):
		return len(value) == 0
	else:
		return False


def is_not_empty(value: Any) -> bool:
	return not is_empty(value)


def equals_decimal(value: Union[int, float, Decimal], another: Any) -> bool:
	if isinstance(value, int) or isinstance(value, float):
		value = Decimal(value)

	if isinstance(another, Decimal):
		return value == another
	elif isinstance(another, int) or isinstance(another, float):
		return value == Decimal(another)
	elif isinstance(another, str):
		parsed, another_value = is_decimal(another)
		return parsed and value == another_value
	else:
		return False


def equals_time(value: time, another: Any, formats: List[str]) -> bool:
	"""
	microsecond is ignored
	"""
	if isinstance(another, time):
		return value.replace(microsecond=0) == another.replace(microsecond=0)
	elif isinstance(another, str):
		parsed, another_value = is_time(another, formats)
		return parsed and value.replace(microsecond=0) == another_value.replace(microsecond=0)
	else:
		return False


def equals_date(value: date, another: Any, formats: List[str]) -> bool:
	"""
	hour, minute, second, microsecond and timezone are ignored
	"""
	value = truncate_time(value)
	if isinstance(another, datetime):
		return value == truncate_time(another)
	if isinstance(another, date):
		return value == truncate_time(another)
	elif isinstance(another, str):
		parsed, another_value = is_date(another, formats)
		if not parsed or another_value is None:
			return False
		else:
			return value == truncate_time(another_value)
	else:
		return False


def equals_bool(value: bool, another: Any) -> bool:
	if isinstance(another, bool):
		return value == another
	elif isinstance(another, (int, float, Decimal)):
		if value:
			return another == 1
		else:
			return another == 0
	elif isinstance(another, str):
		if value:
			return another.lower() in ['1', 't', 'true', 'y', 'yes']
		else:
			return another.lower() in ['0', 'f', 'false', 'n', 'no']
	else:
		return False


def value_equals(
		one: Any, another: Any,
		time_formats: List[str], date_formats: List[str]) -> bool:
	if one is None or (isinstance(one, str) and one == ''):
		# empty string is none
		return another is None or (isinstance(another, str) and another == '')
	elif another is None or (isinstance(another, str) and another == ''):
		return False
	elif isinstance(one, bool):
		return equals_bool(one, another)
	elif isinstance(another, bool):
		return equals_bool(another, one)
	elif isinstance(one, int) or isinstance(one, float) or isinstance(one, Decimal):
		# compare numeric
		return equals_decimal(Decimal(one), another)
	elif isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
		# compare numeric
		return equals_decimal(Decimal(another), one)
	elif isinstance(one, time):
		# compare time
		return equals_time(one, another, time_formats)
	elif isinstance(another, time):
		# compare time
		return equals_time(another, one, time_formats)
	elif isinstance(one, datetime) or isinstance(one, date):
		# compare datetime or date
		return equals_date(one, another, date_formats)
	elif isinstance(another, datetime) or isinstance(another, date):
		# compare datetime or date
		return equals_date(another, one, date_formats)
	elif isinstance(one, str):
		# compare string
		if isinstance(another, int) or isinstance(another, float) or isinstance(another, Decimal):
			return one == str(another)
		elif isinstance(another, str):
			return one == another
		else:
			return False
	else:
		# any other type is not comparable
		return False


def value_not_equals(
		one: Any, another: Any,
		time_formats: List[str], date_formats: List[str]) -> bool:
	return not value_equals(one, another, time_formats, date_formats)


def less_or_equals_decimal(
		value: Union[int, float, Decimal], another: Any, allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	if isinstance(value, int) or isinstance(value, float):
		value = Decimal(value)

	if isinstance(another, Decimal):
		return True, value <= another if allow_equals else value < another
	elif isinstance(another, int) or isinstance(another, float):
		return True, value <= Decimal(another) if allow_equals else value < Decimal(another)
	elif isinstance(another, str):
		parsed, another_value = is_decimal(another)
		if parsed:
			return True, value <= another_value if allow_equals else value < another_value
		else:
			return False, None
	else:
		return False, None


def greater_or_equals_decimal(
		value: Union[int, float, Decimal], another: Any, allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	if isinstance(value, int) or isinstance(value, float):
		value = Decimal(value)

	if isinstance(another, Decimal):
		return True, value >= another if allow_equals else value > another
	elif isinstance(another, int) or isinstance(another, float):
		return True, value >= Decimal(another) if allow_equals else value > Decimal(another)
	elif isinstance(another, str):
		parsed, another_value = is_decimal(another)
		if parsed:
			return True, value >= another_value if allow_equals else value > another_value
		else:
			return False, None
	else:
		return False, None


def less_or_equals_time(
		value: time, another: Any, formats: List[str], allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	if isinstance(another, time):
		return True, value.replace(microsecond=0) <= another.replace(microsecond=0) \
			if allow_equals else value.replace(microsecond=0) < another.replace(microsecond=0)
	elif isinstance(another, str):
		parsed, another_value = is_time(another, formats)
		if parsed:
			return True, value.replace(microsecond=0) <= another_value.replace(microsecond=0) \
				if allow_equals else value.replace(microsecond=0) < another_value.replace(microsecond=0)
		else:
			return False, None
	else:
		return False, None


def greater_or_equals_time(
		value: time, another: Any, formats: List[str], allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	if isinstance(another, time):
		return True, value.replace(microsecond=0) >= another.replace(microsecond=0) \
			if allow_equals else value.replace(microsecond=0) > another.replace(microsecond=0)
	elif isinstance(another, str):
		parsed, another_value = is_time(another, formats)
		if parsed:
			return True, value.replace(microsecond=0) >= another_value.replace(microsecond=0) \
				if allow_equals else value.replace(microsecond=0) > another_value.replace(microsecond=0)
		else:
			return False, None
	else:
		return False, None


def less_or_equals_date(
		value: date, another: Any, formats: List[str], allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	value = truncate_time(value)
	if isinstance(another, datetime):
		return True, value <= truncate_time(another) \
			if allow_equals else value < truncate_time(another)
	elif isinstance(another, str):
		parsed, another_value = is_date(another, formats)
		if parsed:
			return True, value <= truncate_time(another_value) if allow_equals else value < truncate_time(another_value)
		else:
			return False, None
	else:
		return False, None


def greater_or_equals_date(
		value: date, another: Any, formats: List[str], allow_equals: bool
) -> Tuple[bool, Optional[bool]]:
	"""
	return a tuple which first is can be compared.
	second is comparison result when first is true, otherwise second is none.
	"""
	value = truncate_time(value)
	if isinstance(another, datetime):
		return True, value >= truncate_time(another) \
			if allow_equals else value > truncate_time(another)
	elif isinstance(another, str):
		parsed, another_value = is_date(another, formats)
		if parsed:
			return True, value >= truncate_time(another_value) if allow_equals else value > truncate_time(another_value)
		else:
			return False, None
	else:
		return False, None



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from re import sub
from typing import Any, List, Optional, Tuple, Union

from .array_helper import ArrayHelper


class DateTimeEncoder(JSONEncoder):
	def default(self, o):
		if isinstance(o, (datetime, date, time)):
			return o.isoformat()
		if isinstance(o, Decimal):
			if o == o.to_integral_value():
				# integral
				if o > 999999999999999 or o < -999999999999999:
					return f'{int(o)}'
				else:
					return int(o)
			elif o > 999999999999999 or o < -999999999999999:
				return f'{o}'
			else:
				return float(o)
		return super().default(o)


def get_current_time_in_seconds() -> datetime:
	return datetime.now().replace(tzinfo=None, microsecond=0)


def is_date_or_time_instance(value: Any) -> bool:
	return value is not None and (isinstance(value, date) or isinstance(value, time) or isinstance(value, datetime))


def truncate_time(value: Union[date, datetime]) -> datetime:
	if isinstance(value, datetime):
		return value.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
	else:
		return datetime(year=value.year, month=value.month, day=value.day) \
			.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)


def last_day_of_month(a_date: date) -> int:
	return ((a_date.replace(day=1) + timedelta(days=31)).replace(day=1) - timedelta(days=1)).day


def year_diff(end_date: date, start_date: date) -> int:
	end_year = end_date.year
	end_month = end_date.month
	end_day = end_date.day
	start_year = start_date.year
	start_month = start_date.month
	start_day = start_date.day
	if end_year == start_year:
		# same year, always return 0
		return 0
	elif end_year > start_year:
		if end_month == start_month:
			if end_day >= start_day:
				return end_year - start_year
			elif end_month == 2:
				last_day_of_end_month = last_day_of_month(end_date)
				if end_day == last_day_of_end_month and start_day >= end_day:
					return end_year - start_year
				else:
					return end_year - start_year - 1
			else:
				return end_year - start_year - 1
		elif end_month > start_month:
			return end_year - start_year
		else:
			return end_year - start_year - 1
	else:
		if end_month == start_month:
			if end_day > start_day:
				if end_month == 2:
					last_day_of_start_month = last_day_of_month(start_date)
					if start_day == last_day_of_start_month:
						return end_year - start_year
					else:
						return end_year - start_year + 1
				else:
					return end_year - start_year + 1
			else:
				return end_year - start_year
		elif end_month > start_month:
			return end_year - start_year + 1
		else:
			return end_year - start_year


def month_diff(end_date: date, start_date: date) -> int:
	end_year = end_date.year
	end_month = end_date.month
	end_day = end_date.day
	start_year = start_date.year
	start_month = start_date.month
	start_day = start_date.day
	if end_year == start_year:
		if end_month == start_month:
			# same year, same month, always return 0
			return 0
		if end_month > start_month:
			if end_day >= start_day:
				return end_month - start_month
			else:
				last_day_of_end_month = last_day_of_month(end_date)
				if last_day_of_end_month == end_day and start_day >= end_day:
					# it is last day of end month
					return end_month - start_month
				else:
					return end_month - start_month - 1
		else:
			# end date is before start date
			if end_day > start_day:
				last_day_of_start_month = last_day_of_month(start_date)
				if last_day_of_start_month == start_day and end_day >= start_day:
					# it is last day of start month
					return end_month - start_month
				else:
					return end_month - start_month + 1
			else:
				return end_month - start_month
	elif end_year > start_year:
		if end_day >= start_day:
			return (end_year - start_year) * 12 + end_month - start_month
		else:
			last_day_of_end_month = last_day_of_month(end_date)
			if last_day_of_end_month == end_day and start_day >= end_day:
				return (end_year - start_year) * 12 + end_month - start_month
			else:
				return (end_year - start_year) * 12 + end_month - start_month + 1
	else:
		# end year is before start year
		if end_day > start_day:
			last_day_of_start_month = last_day_of_month(start_date)
			if last_day_of_start_month == start_day and end_day >= start_day:
				# it is last day of start month
				return (end_year - start_year + 1) * 12 + 12 - end_month + start_month
			else:
				return (end_year - start_year + 1) * 12 + 12 - end_month + start_month - 1
		else:
			return (end_year - start_year + 1) * 12 + 12 - end_month + start_month


def try_to_format_time(might_be_time: str, time_format: str) -> Tuple[bool, Optional[time]]:
	"""
	return a datetime object is parsed
	"""
	try:
		d = datetime.strptime(might_be_time, time_format)
		return True, d.time()
	except ValueError:
		return False, None


def is_time(value: Optional[str], formats: List[str]) -> Tuple[bool, Optional[time]]:
	"""
	none is not a time value, otherwise remove non-number characters and try to parse by given formats.
	digits after removing must match digits of format
	"""
	if value is None:
		return False, None
	tidy_value = sub(r'[^0-9+]', '', value)
	count = len(tidy_value)
	suitable_formats = ArrayHelper(formats).filter(lambda x: len(x) == count).to_list()
	for suitable_format in suitable_formats:
		parsed, date_value = try_to_format_time(tidy_value, suitable_format)
		if parsed:
			return parsed, date_value
	return False, None


def try_to_time(value: Any, formats: List[str]) -> Optional[time]:
	"""
	try to parse given value to date, or returns none when cannot be parsed.
	formats can be datetime and date format
	"""
	if value is None:
		return None
	elif isinstance(value, time):
		return value
	elif isinstance(value, str):
		parsed, date_value = is_time(value, formats)
		if parsed:
			return date_value
	return None


def try_to_format_date(might_be_date: str, date_format: str) -> Tuple[bool, Optional[date]]:
	"""
	return a datetime object is parsed
	"""
	try:
		d = datetime.strptime(might_be_date, date_format)
		return True, d
	except ValueError:
		return False, None


def is_suitable_format(value_length: int, a_format: str) -> bool:
	plus_year = '%Y' in a_format
	plus_timezone = '%z' in a_format
	plus_digits = (2 if plus_year else 0) + (3 if plus_timezone else 0)
	if value_length > 14 and not plus_timezone:
		return '%f' in a_format
	elif value_length > 14 and plus_timezone:
		if value_length > 19:
			return '%f' in a_format and plus_timezone
		else:
			return value_length == len(a_format) + plus_digits
	else:
		return value_length == len(a_format) + plus_digits


def is_date_plus_format(value: Optional[str], formats: List[str]) -> Tuple[bool, Optional[date], Optional[str]]:
	"""
	none is not a date value, otherwise remove non-number characters and try to parse by given formats.
	digits after removing must match digits of format.
	return format itself when parsed
	"""
	if value is None:
		return False, None, None

	tidy_value = sub(r'[^0-9+]', '', value)
	count = len(tidy_value)
	# format cannot use length to match
	suitable_formats = ArrayHelper(formats).filter(lambda x: is_suitable_format(count, x)).to_list()
	for suitable_format in suitable_formats:
		parsed, date_value = try_to_format_date(tidy_value, suitable_format)
		if parsed:
			return parsed, date_value, suitable_format
	return False, None, None


def is_date(value: Optional[str], formats: List[str]) -> Tuple[bool, Optional[date]]:
	parsed, date_value, _ = is_date_plus_format(value, formats)
	return parsed, date_value


def is_datetime(value: Optional[str], formats: List[str]) -> Tuple[bool, Optional[datetime]]:
	parsed, date_value = is_date(value, formats)
	if not parsed:
		return False, None
	elif isinstance(date_value, datetime):
		return True, date_value
	else:
		return True, datetime(
			year=date_value.year, month=date_value.month, day=date_value.day,
			hour=0, minute=0, second=0, microsecond=0)


def try_to_date(value: Any, formats: List[str], allow_timestamp: bool = False) -> Optional[date]:
	"""
	try to parse given value to date, or returns none when cannot be parsed.
	formats can be datetime and date format
	"""
	if value is None:
		return None
	elif isinstance(value, date):
		return value
	elif allow_timestamp and (isinstance(value, int) or isinstance(value, float)):
		# timestamp
		return datetime.fromtimestamp(value, tz=None)
	elif allow_timestamp and isinstance(value, Decimal):
		return datetime.fromtimestamp(float(value), tz=None)
	elif isinstance(value, str):
		parsed, date_value = is_date(value, formats)
		if parsed:
			return date_value
	return None


DATE_FORMAT_MAPPING = {
	'Y': '%Y',  # 4 digits year
	'y': '%y',  # 2 digits year
	'M': '%m',  # 2 digits month
	'D': '%d',  # 2 digits day of month
	'h': '%H',  # 2 digits hour, 00 - 23
	'H': '%I',  # 2 digits hour, 01 - 12
	'm': '%M',  # 2 digits minute
	's': '%S',  # 2 digits second
	'W': '%A',  # Monday - Sunday
	'w': '%a',  # Mon - Sun
	'B': '%B',  # January - December
	'b': '%b',  # Jan - Dec
	'p': '%p'  # AM/PM
}


def translate_date_format_to_memory(date_format: str) -> str:
	return ArrayHelper(list(DATE_FORMAT_MAPPING)) \
		.reduce(lambda original, x: original.replace(x, DATE_FORMAT_MAPPING[x]), date_format)


class DateTimeConstants(int, Enum):
	"""
	Week starts from Sunday
	"""
	HALF_YEAR_FIRST = 1
	HALF_YEAR_SECOND = 2

	QUARTER_FIRST = 1
	QUARTER_SECOND = 2
	QUARTER_THIRD = 3
	QUARTER_FOURTH = 4

	JANUARY = 1
	FEBRUARY = 2
	MARCH = 3
	APRIL = 4
	MAY = 5
	JUNE = 6
	JULY = 7
	AUGUST = 8
	SEPTEMBER = 9
	OCTOBER = 10
	NOVEMBER = 11
	DECEMBER = 12

	HALF_MONTH_FIRST = 1
	HALF_MONTH_SECOND = 2

	TEN_DAYS_FIRST = 1
	TEN_DAYS_SECOND = 2
	TEN_DAYS_THIRD = 3

	WEEK_OF_YEAR_FIRST_SHORT = 0  # first week less than 7 days, otherwise week of year starts from 1
	WEEK_OF_YEAR_FIRST = 1
	WEEK_OF_YEAR_LAST = 53

	WEEK_OF_MONTH_FIRST_SHORT = 0  # first week less than 7 days, otherwise week of month starts from 1
	WEEK_OF_MONTH_FIRST = 1
	WEEK_OF_MONTH_LAST = 5

	HALF_WEEK_FIRST = 1
	HALF_WEEK_SECOND = 2

	DAY_OF_MONTH_MIN = 1
	DAY_OF_MONTH_MAX = 31

	SUNDAY = 1
	MONDAY = 2
	TUESDAY = 3
	WEDNESDAY = 4
	THURSDAY = 5
	FRIDAY = 6
	SATURDAY = 7

	DAY_KIND_WORKDAY = 1
	DAY_KIND_WEEKEND = 2
	DAY_KIND_HOLIDAY = 3

	HOUR_KIND_WORKTIME = 1
	HOUR_KIND_OFF_HOURS = 2
	HOUR_KIND_SLEEPING_TIME = 3

	AM = 1
	PM = 2


def get_year(dt: date) -> int:
	return dt.year


def get_month(dt: date) -> int:
	return dt.month


def get_half_year(dt: date) -> int:
	return DateTimeConstants.HALF_YEAR_FIRST.value if get_month(dt) <= 6 else DateTimeConstants.HALF_YEAR_SECOND.value


def get_quarter(dt: date) -> int:
	month = get_month(dt)
	if month <= 3:
		return DateTimeConstants.QUARTER_FIRST.value
	elif month <= 6:
		return DateTimeConstants.QUARTER_SECOND.value
	elif month <= 9:
		return DateTimeConstants.QUARTER_THIRD.value
	else:
		return DateTimeConstants.QUARTER_FOURTH.value


def get_week_of_year(dt: date) -> int:
	return int(dt.strftime('%U'))


def get_week_of_month(dt: date) -> int:
	first_day = dt.replace(day=1)
	first_day_week = get_week_of_year(first_day)
	week_of_year = get_week_of_year(dt)
	if first_day_week == week_of_year:
		if get_day_of_week(first_day) == DateTimeConstants.SUNDAY.value:
			# first week is full week
			return DateTimeConstants.WEEK_OF_MONTH_FIRST
		else:
			# first week is short
			return DateTimeConstants.WEEK_OF_MONTH_FIRST_SHORT
	else:
		if get_day_of_week(first_day) == DateTimeConstants.SUNDAY.value:
			# first week is full week, must add 1
			return week_of_year - first_day_week + 1
		else:
			# first week is short
			return week_of_year - first_day_week


def get_day_of_month(dt: date) -> int:
	return dt.day


def get_day_of_week(dt: date) -> int:
	# iso weekday: Monday is 1 and Sunday is 7
	return (dt.isoweekday() + 1) % 8


def to_start_of_day(process_date: date):
	if isinstance(process_date, datetime):
		return process_date.date()
	else:
		return process_date


def to_yesterday(process_date: date) -> date:
	"""
	return yesterday
	"""
	return to_start_of_day(process_date - timedelta(days=1))


def to_previous_week(process_date: date) -> date:
	"""
	return first day of previous week, sunday
	"""
	# iso weekday: Monday is 1 and Sunday is 7
	weekday = process_date.isoweekday()
	# get last sunday
	return to_start_of_day(process_date - timedelta(days=weekday % 7 + 7))


def to_previous_month(process_date: date) -> date:
	"""
	return first day of previous month
	"""
	# get last day of previous month
	process_date = process_date.replace(day=1) - timedelta(days=1)
	# set to first day of previous month
	return to_start_of_day(process_date.replace(day=1))



from decimal import Decimal
from typing import Any, Optional, Tuple


def is_numeric_instance(value: Any) -> bool:
	return value is not None and (isinstance(value, int) or isinstance(value, float) or isinstance(value, Decimal))


def is_decimal(value: Optional[str]) -> Tuple[bool, Optional[Decimal]]:
	"""
	none is not a numeric value, otherwise try to parse it by float function
	"""
	if value is None:
		return False, None
	try:
		return True, Decimal(value)
	except ValueError:
		return False, None


def try_to_decimal(value: Any) -> Optional[Decimal]:
	"""
	try to parse given value to decimal, or returns none when cannot be parsed
	"""
	if value is None:
		return None
	elif isinstance(value, int) or isinstance(value, float):
		return Decimal(value)
	elif isinstance(value, Decimal):
		return value
	elif isinstance(value, str):
		parsed, decimal_value = is_decimal(value)
		if parsed:
			return decimal_value
	return None
